@@ -1,4 +1,4 @@
-//  Copyright (C) 2010 Ben Asselstine
+//  Copyright (C) 2010, 2011 Ben Asselstine
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -32,7 +32,7 @@ ChangesEditorDialog * ChangesEditorDialog::create(Tapioca::Session *s)
 				    + "/changes-editor-dialog.gtk");
 
   ChangesEditorDialog *dialog;
-  xml->get_widget("dialog", dialog);
+  xml->get_widget_derived("dialog", dialog);
   dialog->on_dialog_loaded(s);
   return dialog;
 }
@@ -192,74 +192,85 @@ void ChangesEditorDialog::update_buttons()
 
 void ChangesEditorDialog::fill_modified_galleries(std::list<Tapioca::ChangedGallery> &l)
 {
-  modified_galleries_label->set_text 
-    (String::ucompose("Modified Galleries (%1)", l.size()));
   modified_galleries_list->clear();
-  modified_galleries_expander->set_expanded(l.size() != 0);
-  modified_galleries_expander->set_sensitive(l.size() != 0);
   for (std::list<Tapioca::ChangedGallery>::iterator i = l.begin(); 
        i != l.end(); i++)
     add_row_modified_gallery(&(*i));
+  size_t num = modified_galleries_list->children().size();
+  modified_galleries_label->set_text 
+    (String::ucompose("Modified Galleries (%1)", num));
+  modified_galleries_expander->set_expanded(num != 0);
+  modified_galleries_expander->set_sensitive(num != 0);
 }
 
 void ChangesEditorDialog::fill_modified_documents(std::list<Tapioca::ChangedDocument> &l)
 {
-  modified_documents_label->set_text 
-    (String::ucompose("Modified Documents (%1)", l.size()));
   modified_documents_list->clear();
-  modified_documents_expander->set_expanded(l.size() != 0);
-  modified_documents_expander->set_sensitive(l.size() != 0);
   for (std::list<Tapioca::ChangedDocument>::iterator i = l.begin(); 
        i != l.end(); i++)
     add_row_modified_document(&(*i));
+  size_t num = modified_documents_list->children().size();
+  modified_documents_label->set_text 
+    (String::ucompose("Modified Documents (%1)", num));
+  modified_documents_expander->set_expanded(num != 0);
+  modified_documents_expander->set_sensitive(num != 0);
 }
 
 void ChangesEditorDialog::fill_removed_galleries(std::list<Tapioca::ChangedGallery> &l)
 {
-  removed_galleries_label->set_text 
-    (String::ucompose("Removed Galleries (%1)", l.size()));
   removed_galleries_list->clear();
-  removed_galleries_expander->set_expanded(l.size() != 0);
-  removed_galleries_expander->set_sensitive(l.size() != 0);
   for (std::list<Tapioca::ChangedGallery>::iterator i = l.begin(); 
        i != l.end(); i++)
     add_row_removed_gallery(&(*i));
+  size_t num = removed_galleries_list->children().size();
+  removed_galleries_label->set_text 
+    (String::ucompose("Removed Galleries (%1)", num));
+  removed_galleries_expander->set_expanded(num != 0);
+  removed_galleries_expander->set_sensitive(num != 0);
 }
 
 void ChangesEditorDialog::fill_added_galleries(std::list<Tapioca::ChangedGallery> &l)
 {
-  added_galleries_label->set_text 
-    (String::ucompose("Added Galleries (%1)", l.size()));
   added_galleries_list->clear();
-  added_galleries_expander->set_expanded(l.size() != 0);
-  added_galleries_expander->set_sensitive(l.size() != 0);
   for (std::list<Tapioca::ChangedGallery>::iterator i = l.begin(); 
        i != l.end(); i++)
     add_row_added_gallery(&(*i));
+  size_t num = added_galleries_list->children().size();
+  added_galleries_label->set_text 
+    (String::ucompose("Added Galleries (%1)", num));
+  added_galleries_expander->set_expanded(num != 0);
+  added_galleries_expander->set_sensitive(num != 0);
 }
 
 void ChangesEditorDialog::fill_removed_documents(std::list<Tapioca::ChangedDocument> &l)
 {
-  removed_documents_label->set_text 
-    (String::ucompose("Removed Documents (%1)", l.size()));
   removed_documents_list->clear();
-  removed_documents_expander->set_expanded(l.size() != 0);
-  removed_documents_expander->set_sensitive(l.size() != 0);
   for (std::list<Tapioca::ChangedDocument>::iterator i = l.begin(); 
        i != l.end(); i++)
-    add_row_removed_document(&(*i));
+    {
+      //only add it if we're not removing the gallery.
+      //Gallery *parent = session->find_by_id((*i).first.get_gallery_id());
+      //if (parent)
+        add_row_removed_document(&(*i));
+    }
+  size_t num = removed_documents_list->children().size();
+  removed_documents_expander->set_expanded(num != 0);
+  removed_documents_expander->set_sensitive(num != 0);
+  removed_documents_label->set_text 
+    (String::ucompose("Removed Documents (%1)", num));
 }
 
 void ChangesEditorDialog::fill_added_documents(std::list<Tapioca::ChangedDocument> &l)
 {
-  added_documents_label->set_text 
-    (String::ucompose("Added Documents (%1)", l.size()));
   added_documents_list->clear();
-  added_documents_expander->set_expanded(l.size() != 0);
-  added_documents_expander->set_sensitive(l.size() != 0);
   for (std::list<Tapioca::ChangedDocument>::iterator i = l.begin(); 
        i != l.end(); i++)
     add_row_added_document(&(*i));
+  size_t num = added_documents_list->children().size();
+  added_documents_label->set_text 
+    (String::ucompose("Added Documents (%1)", num));
+  added_documents_expander->set_expanded(num != 0);
+  added_documents_expander->set_sensitive(num != 0);
 }
 
 void ChangesEditorDialog::add_row_modified_gallery(Tapioca::ChangedGallery *gallery)
@@ -328,7 +339,8 @@ void ChangesEditorDialog::on_modified_gallery_activated(const Gtk::TreeModel::Pa
 {
   Gallery *gallery = new Gallery(get_selected_modified_gallery()->first);
   GalleryDialog *d = GalleryDialog::create
-    (gallery, &get_selected_modified_gallery()->second);
+    (gallery, &get_selected_modified_gallery()->second, 
+     session->get_tags());
   d->set_transient_for(*this);
   int response = d->run();
   d->hide();
@@ -351,7 +363,9 @@ void ChangesEditorDialog::on_modified_document_activated(const Gtk::TreeModel::P
 {
   Document *document = new Document(get_selected_modified_document()->first);
   DocumentDialog *d = DocumentDialog::create
-    (document, &get_selected_modified_document()->second);
+    (document, &get_selected_modified_document()->second,
+     session->get_tags());
+  d->view_document.connect(sigc::mem_fun(view_document, &sigc::signal<void, Tapioca::Document*>::emit));
   d->set_transient_for(*this);
   int response = d->run();
   d->hide();
@@ -374,7 +388,7 @@ void ChangesEditorDialog::on_removed_gallery_activated(const Gtk::TreeModel::Pat
 {
   ChangedGallery * gallery= get_selected_removed_gallery();
   GalleryDialog *d = GalleryDialog::create (&gallery->first, &gallery->second, 
-                                            true);
+                                            session->get_tags(), true);
   d->set_transient_for(*this);
   d->run();
   delete d;
@@ -385,7 +399,7 @@ void ChangesEditorDialog::on_added_gallery_activated(const Gtk::TreeModel::Path&
 {
   Gallery *gallery = new Gallery(get_selected_added_gallery()->first);
   GalleryDialog* d = GalleryDialog::create
-    (gallery, &get_selected_added_gallery()->second);
+    (gallery, &get_selected_added_gallery()->second, session->get_tags());
   d->set_transient_for(*this);
   int response = d->run();
   d->hide();
@@ -408,7 +422,9 @@ void ChangesEditorDialog::on_removed_document_activated(const Gtk::TreeModel::Pa
 {
   ChangedDocument* document = get_selected_removed_document();
   DocumentDialog *d = 
-    DocumentDialog::create(&document->first, &document->second, true);
+    DocumentDialog::create(&document->first, &document->second, 
+                           session->get_tags(), true);
+  d->view_document.connect(sigc::mem_fun(view_document, &sigc::signal<void, Tapioca::Document*>::emit));
   d->set_transient_for(*this);
   d->run();
   delete d;
@@ -419,7 +435,9 @@ void ChangesEditorDialog::on_added_document_activated(const Gtk::TreeModel::Path
 {
   Document *document = new Document(get_selected_added_document()->first);
   DocumentDialog *d = DocumentDialog::create
-    (document, &get_selected_added_document()->second);
+    (document, &get_selected_added_document()->second,
+     session->get_tags());
+  d->view_document.connect(sigc::mem_fun(view_document, &sigc::signal<void, Tapioca::Document*>::emit));
   d->set_transient_for(*this);
   int response = d->run();
   d->hide();
