@@ -1,4 +1,4 @@
-//  Copyright (C) 2010 Ben Asselstine
+//  Copyright (C) 2010, 2011 Ben Asselstine
 //
 //  This program is free software; you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -21,27 +21,34 @@
 #include "main.h"
 #include "ucompose.hpp"
 #include <cstring>
+#include "tapioca-window.h"
 
 std::string ItemTaggingWidget::default_tag_name = "type your tag";
 ItemTaggingWidget::ItemTaggingWidget()
-    : name_column("Name", name_renderer)
+    : name_column("Tags", name_renderer), checked_column("", checked_renderer)
 {
   inhibit = false;
   Glib::RefPtr<Gtk::Builder> xml
     = Gtk::Builder::create_from_file(Main::get_glade_path()
                                      + "/item-tagging-dialog.gtk");
-
-
   Gtk::Dialog *d;
   xml->get_widget("dialog", d);
   xml->get_widget("contents", contents);
-  xml->get_widget("tags_count_label", tags_count_label);
   xml->get_widget("treeview", treeview);
   tags_list = Gtk::ListStore::create(tags_columns);
   treeview->set_model(tags_list);
   treeview->get_selection()->set_mode(Gtk::SELECTION_SINGLE);
   treeview->get_selection()->signal_changed().connect
     (sigc::mem_fun(*this, &ItemTaggingWidget::on_tag_selected));
+  //the checked toggle
+  checked_renderer.property_mode() = Gtk::CELL_RENDERER_MODE_EDITABLE;
+  checked_renderer.property_activatable() = true;
+  checked_renderer.signal_editing_started().connect
+    (sigc::mem_fun(*this, &ItemTaggingWidget::on_checked_changed));
+  checked_column.set_cell_data_func
+    (checked_renderer, sigc::mem_fun(*this, &ItemTaggingWidget::cell_data_checked));
+  treeview->append_column(checked_column);
+  //the name entry box
   name_renderer.property_editable() = true;
   name_renderer.signal_edited().connect
     (sigc::mem_fun(*this, &ItemTaggingWidget::on_tag_edited));
@@ -50,31 +57,31 @@ ItemTaggingWidget::ItemTaggingWidget()
   name_column.set_cell_data_func
     (name_renderer, sigc::mem_fun(*this, &ItemTaggingWidget::cell_data_tag));
   treeview->append_column(name_column);
+
   xml->get_widget("add_button", add_button);
   add_button->signal_clicked().connect
     (sigc::mem_fun(this,&ItemTaggingWidget::on_add_clicked));
-  xml->get_widget("remove_button", remove_button);
-  remove_button->signal_clicked().connect
-    (sigc::mem_fun(this,&ItemTaggingWidget::on_remove_clicked));
 }
 
 ItemTaggingWidget::~ItemTaggingWidget()
 {
 }
 
-void ItemTaggingWidget::fill_tags(std::list<std::string> t)
+void ItemTaggingWidget::fill_tags(std::list<std::string> t, std::list<std::string> library)
 {
+  tags_list->clear();
   tags = t;
-  for (std::list<std::string>::iterator it = tags.begin(); 
-       it != tags.end(); it++)
-    add_tag_row(*it);
-  tags_count_label->set_text(String::ucompose("%1", tags.size()));
+  library_tags = library;
+  for (std::list<std::string>::iterator it = library_tags.begin(); 
+       it != library_tags.end(); it++)
+    add_tag_row(*it, has_tag(*it));
 }
 
-void ItemTaggingWidget::add_tag_row(std::string tag)
+void ItemTaggingWidget::add_tag_row(std::string tag, bool checked)
 {
   Gtk::TreeIter i = tags_list->append();
   (*i)[tags_columns.name] = tag;
+  (*i)[tags_columns.checked] = checked;
 }
 
 std::string ItemTaggingWidget::get_selected_tag()
@@ -90,6 +97,12 @@ std::string ItemTaggingWidget::get_selected_tag()
   return "";
 }
 
+bool ItemTaggingWidget::has_library_tag(std::string tag) const
+{
+  return (std::find(library_tags.begin(), library_tags.end(), tag) != 
+          library_tags.end());
+}
+
 bool ItemTaggingWidget::has_tag(std::string tag) const
 {
   return (std::find(tags.begin(), tags.end(), tag) != tags.end());
@@ -126,6 +139,7 @@ void ItemTaggingWidget::on_add_clicked()
     {
       Gtk::TreeIter i = tags_list->append();
       (*i)[tags_columns.name] = tag;
+      (*i)[tags_columns.checked] = true;
       Glib::RefPtr<Gtk::TreeSelection> selection = treeview->get_selection();
       selection->select(i);
     }
@@ -134,27 +148,10 @@ void ItemTaggingWidget::on_add_clicked()
   treeview->get_cursor(path, column);
   path = Gtk::TreeModel::Path(String::ucompose("%1", tags_list->children().size()-1));
   treeview->set_cursor(path, name_column, true);
-  tags_count_label->set_text(String::ucompose("%1", tags.size()));
   treeview->set_cursor(path, name_column, true);
   inhibit = false;
 }
 
-void ItemTaggingWidget::on_remove_clicked()
-{
-  Glib::RefPtr<Gtk::TreeSelection> selection = treeview->get_selection();
-  Gtk::TreeModel::iterator iterrow = selection->get_selected();
-
-  if (iterrow) 
-    {
-      Gtk::TreeModel::Row row = *iterrow;
-      Glib::ustring tag = row[tags_columns.name];
-      if (std::find(tags.begin(), tags.end(), tag) != tags.end())
-        tags.remove(tag);
-      tags_list->erase(iterrow);
-      tags_count_label->set_text(String::ucompose("%1", tags.size()));
-    }
-}
-
 void
 ItemTaggingWidget::cell_data_tag(Gtk::CellRenderer *renderer, const Gtk::TreeIter &i)
 {
@@ -168,6 +165,8 @@ ItemTaggingWidget::on_tag_edited(const Glib::ustring &path,
 {
   Gtk::TreeModel::iterator iterrow = tags_list->get_iter(Gtk::TreePath(path));
   Gtk::TreeModel::Row row = *iterrow;
+  if (row[tags_columns.checked] == false)
+    return;
   Glib::ustring old_tag = row[tags_columns.name];
   row[tags_columns.name] = new_text;
   if (validate_tag(new_text) == true)
@@ -177,11 +176,14 @@ ItemTaggingWidget::on_tag_edited(const Glib::ustring &path,
       Gtk::TreeViewColumn *column = NULL;
       Gtk::TreeModel::Path path;
       treeview->get_cursor(path, column);
-      path = Gtk::TreeModel::Path(String::ucompose("%1", tags_list->children().size()-1));
+      path = Gtk::TreeModel::Path
+        (String::ucompose("%1", tags_list->children().size()-1));
       treeview->get_selection()->select(path);
-      on_remove_clicked();
+      if (old_tag == default_tag_name)
+        on_remove_clicked();
+      else
+        row[tags_columns.name] = old_tag;
     }
-  tags_count_label->set_text(String::ucompose("%1", tags.size()));
 }
 
 bool ItemTaggingWidget::validate_tag(std::string tag, bool show_error)
@@ -219,7 +221,10 @@ void ItemTaggingWidget::on_cancel_tag_edit()
 void ItemTaggingWidget::add_tag(std::string tag)
 {
   if (std::find(tags.begin(), tags.end(), tag) == tags.end())
-    tags.push_back(tag);
+    {
+      tags.push_back(tag);
+      tags_changed.emit(tags);
+    }
 }
 
 void ItemTaggingWidget::update_tag(std::string oldvalue, std::string newvalue)
@@ -230,12 +235,16 @@ void ItemTaggingWidget::update_tag(std::string oldvalue, std::string newvalue)
     (*i) = newvalue;
   else
     tags.push_back(newvalue);
+  tags_changed.emit(tags);
 }
 
 void ItemTaggingWidget::remove_tag(std::string tag)
 {
   if (std::find(tags.begin(), tags.end(), tag) != tags.end())
-    tags.remove(tag);
+    {
+      tags.remove(tag);
+      tags_changed.emit(tags);
+    }
 }
 
 std::list<std::string> ItemTaggingWidget::get_tags()
@@ -244,3 +253,37 @@ std::list<std::string> ItemTaggingWidget::get_tags()
   return tags;
 }
 
+void ItemTaggingWidget::cell_data_checked(Gtk::CellRenderer *renderer,
+					const Gtk::TreeIter& i)
+{
+  dynamic_cast<Gtk::CellRendererToggle*>(renderer)->set_active((*i)[tags_columns.checked]);
+}
+
+void ItemTaggingWidget::on_checked_changed(Gtk::CellEditable *editable,
+                                           const Glib::ustring &path)
+{
+  Gtk::TreeModel::iterator iter = treeview->get_model()->get_iter(path);
+
+  (*iter)[tags_columns.checked] =!(*iter)[tags_columns.checked];
+  bool checked = (*iter)[tags_columns.checked];
+  Glib::ustring tag = (*iter)[tags_columns.name];
+  if (checked)
+    add_tag(tag);
+  else
+    remove_tag(tag);
+}
+
+void ItemTaggingWidget::on_remove_clicked()
+{
+  Glib::RefPtr<Gtk::TreeSelection> selection = treeview->get_selection();
+  Gtk::TreeModel::iterator iterrow = selection->get_selected();
+
+  if (iterrow) 
+    {
+      Gtk::TreeModel::Row row = *iterrow;
+      Glib::ustring tag = row[tags_columns.name];
+      if (std::find(tags.begin(), tags.end(), tag) != tags.end())
+        tags.remove(tag);
+      tags_list->erase(iterrow);
+    }
+}
