@@ -144,14 +144,22 @@ TapiocaWindow::TapiocaWindow(BaseObjectType* baseObject,
   //menubar callbacks
   xml->get_widget("connect_menuitem", connect_menuitem);
   connect_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_connect_clicked));
+  xml->get_widget("connect_toolbutton", connect_toolbutton);
+  connect_toolbutton->signal_clicked().connect(sigc::mem_fun(*this,&TapiocaWindow::on_connect_clicked));
   xml->get_widget("open_offline_menuitem", open_offline_menuitem);
   open_offline_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_open_offline_changes_clicked));
+  xml->get_widget("open_toolbutton", open_toolbutton);
+  open_toolbutton->signal_clicked().connect(sigc::mem_fun(*this,&TapiocaWindow::on_open_offline_changes_clicked));
   xml->get_widget("save_offline_menuitem", save_offline_menuitem);
   save_offline_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_save_offline_changes_clicked));
+  xml->get_widget("save_toolbutton", save_toolbutton);
+  save_toolbutton->signal_clicked().connect(sigc::mem_fun(*this,&TapiocaWindow::on_save_offline_changes_clicked));
   xml->get_widget("save_offline_as_menuitem", save_offline_as_menuitem);
   save_offline_as_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_save_offline_changes_as_clicked));
   xml->get_widget("save_changes_menuitem", save_changes_menuitem);
   save_changes_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_save_changes_to_atpic_clicked));
+  xml->get_widget("publish_toolbutton", publish_toolbutton);
+  publish_toolbutton->signal_clicked().connect(sigc::mem_fun(*this,&TapiocaWindow::on_save_changes_to_atpic_clicked));
   xml->get_widget("export_menuitem", export_menuitem);
   export_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_export_clicked));
   xml->get_widget("exit_menuitem", exit_menuitem);
@@ -163,7 +171,24 @@ TapiocaWindow::TapiocaWindow(BaseObjectType* baseObject,
   image_properties_menuitem->set_sensitive(false);
   xml->get_widget("gallery_properties_menuitem", gallery_properties_menuitem);
   gallery_properties_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_gallery_properties_clicked));
+  xml->get_widget("add_document_menuitem", add_document_menuitem);
+  add_document_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_add_document_clicked));
+  xml->get_widget("remove_document_menuitem", remove_document_menuitem);
+  remove_document_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_remove_document_clicked));
+  xml->get_widget("add_gallery_menuitem", add_gallery_menuitem);
+  add_gallery_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_add_gallery_clicked));
+  xml->get_widget("remove_gallery_menuitem", remove_gallery_menuitem);
+  remove_gallery_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_remove_gallery_clicked));
   gallery_properties_menuitem->set_sensitive(false);
+  xml->get_widget("fullscreen_menuitem", fullscreen_menuitem);
+  fullscreen_menuitem->signal_toggled().connect
+    (sigc::mem_fun(*this, &TapiocaWindow::on_fullscreen_toggled));
+  xml->get_widget("show_tags_menuitem", show_tags_menuitem);
+  show_tags_menuitem->signal_toggled().connect
+    (sigc::mem_fun(*this, &TapiocaWindow::on_show_tags_toggled));
+  xml->get_widget("view_document_menuitem", view_document_menuitem);
+  view_document_menuitem->signal_activate().connect
+    (sigc::mem_fun(*this, &TapiocaWindow::on_view_document_clicked));
   xml->get_widget("about_menuitem", about_menuitem);
   about_menuitem->signal_activate().connect(sigc::mem_fun(*this,&TapiocaWindow::on_about_clicked));
 }
@@ -185,6 +210,7 @@ void TapiocaWindow::on_window_loaded(std::string filename)
   Document::remove_file(temporary_directory);
   Document::create_dir(temporary_directory);
   progress_aborted = false;
+  item_tagging_widget = NULL;
 }
 
 TapiocaWindow::~TapiocaWindow()
@@ -331,6 +357,7 @@ void TapiocaWindow::load_session_in_the_background(ProgressDialog &d)
       boost::archive::text_iarchive ia(ifs);
       d.progress(3.0/6.0);
       Gallery::loading.connect(sigc::mem_fun(d, &ProgressDialog::pulse));
+      Document::loading.connect(sigc::mem_fun(d, &ProgressDialog::pulse));
       ia >> session;
     }
   catch (std::exception &ex)
@@ -424,6 +451,7 @@ void TapiocaWindow::save_session_in_the_background(ProgressDialog &d)
   std::ofstream ofs(current_save_filename.c_str());
   boost::archive::text_oarchive oa(ofs);
   Gallery::saving.connect(sigc::mem_fun(d, &ProgressDialog::pulse));
+  Document::saving.connect(sigc::mem_fun(d, &ProgressDialog::pulse));
   oa << session;
   d.hide();
 }
@@ -475,6 +503,7 @@ void TapiocaWindow::on_save_offline_changes_as_clicked()
 void TapiocaWindow::on_save_changes_to_atpic_clicked()
 {
   ChangesEditorDialog *d = ChangesEditorDialog::create(session);
+  d->view_document.connect(sigc::mem_fun(this, &TapiocaWindow::view_document));
   d->set_transient_for(*this);
   int response = d->run();
   if (response == Gtk::RESPONSE_ACCEPT)
@@ -498,7 +527,13 @@ void TapiocaWindow::on_save_changes_to_atpic_clicked()
           di->hide();
           if (resp == Gtk::RESPONSE_ACCEPT)
             {
-              publish(changes);
+              ProgressDialog pd;
+              pd.set_parent_window(*this);
+              pd.aborted.connect
+                (sigc::mem_fun(*this, &TapiocaWindow::on_progress_aborted));
+              Glib::signal_idle().connect_once(sigc::bind<SessionChanges&, ProgressDialog&>(sigc::mem_fun(*this, &TapiocaWindow::publish), changes, pd));
+              pd.run();
+              //publish(changes);
               //publish these changes now.
             }
           delete di;
@@ -507,10 +542,13 @@ void TapiocaWindow::on_save_changes_to_atpic_clicked()
   delete d;
 }
 
-void TapiocaWindow::on_published_gallery_removed(Gallery *gallery, sigc::slot<void> slot)
+void TapiocaWindow::on_published_gallery_removed(Gallery *gallery, sigc::slot<void> slot, ProgressDialog &d)
 {
+  d.pulse();
   publish_changes.removed_galleries.remove
     (publish_changes.removed_galleries.front());
+  d.set_text(String::ucompose("%1 galleries left to remove", 
+                              publish_changes.removed_galleries.size()));
   std::list<ChangedDocument> shadowed_docs;
   for (std::list<ChangedDocument>::iterator i = 
        publish_changes.removed_documents.begin(); 
@@ -525,17 +563,20 @@ void TapiocaWindow::on_published_gallery_removed(Gallery *gallery, sigc::slot<vo
   session->get_profile()->remove(gallery);
   delete gallery;
   if (publish_changes.removed_galleries.size() > 0)
-    remove_published_galleries(slot);
+    remove_published_galleries(slot, d);
   else
     slot();
 }
 
-void TapiocaWindow::on_published_gallery_added(Gallery *gallery, sigc::slot<void> slot)
+void TapiocaWindow::on_published_gallery_added(Gallery *gallery, sigc::slot<void> slot, ProgressDialog &d)
 {
+  d.pulse();
   publish_changes.added_galleries.remove
     (publish_changes.added_galleries.front());
+  d.set_text(String::ucompose("%1 galleries left to add", 
+                              publish_changes.added_galleries.size()));
   if (publish_changes.added_galleries.size() > 0)
-    add_published_galleries(slot);
+    add_published_galleries(slot, d);
   else
     slot();
 }
@@ -659,7 +700,10 @@ void TapiocaWindow::on_image_properties_clicked()
 {
   Document *document = new Document(*get_selected_document());
   DocumentDialog *d = DocumentDialog::create
-    (document, session->get_profile()->find_doc_by_id(document->get_id()));
+    (document, session->get_profile()->find_doc_by_id(document->get_id()),
+     session->get_tags());
+  d->tags_changed.connect(sigc::bind(sigc::mem_fun(this, &TapiocaWindow::on_document_tags_changed), document));
+  d->view_document.connect(sigc::mem_fun(this, &TapiocaWindow::view_document));
   d->set_transient_for(*this);
   int response = d->run();
   d->hide();
@@ -683,7 +727,9 @@ void TapiocaWindow::on_gallery_properties_clicked()
 {
   Gallery *gallery = new Gallery(*get_selected_gallery());
   GalleryDialog *d = GalleryDialog::create
-    (gallery, session->get_profile()->find_by_id(gallery->get_id()));
+    (gallery, session->get_profile()->find_by_id(gallery->get_id()),
+     session->get_tags());
+  d->tags_changed.connect(sigc::bind(sigc::mem_fun(this, &TapiocaWindow::on_gallery_tags_changed), gallery));
 
   d->set_transient_for(*this);
   int response = d->run();
@@ -785,18 +831,41 @@ Document* TapiocaWindow::get_selected_document()
 
 void TapiocaWindow::update_menuitems()
 {
+  open_toolbutton->set_sensitive(true);
+  open_offline_menuitem->set_sensitive(true);
+  connect_toolbutton->set_sensitive(true);
+  connect_menuitem->set_sensitive(true);
   Gallery *gallery = get_selected_gallery();
   if (session != NULL)
     {
       save_offline_menuitem->set_sensitive(true);
+      save_toolbutton->set_sensitive(true);
       save_offline_as_menuitem->set_sensitive(true);
       save_changes_menuitem->set_sensitive(true);
+      publish_toolbutton->set_sensitive(true);
       export_menuitem->set_sensitive(true);
       contents->set_sensitive(true);
       if (gallery)
-        gallery_properties_menuitem->set_sensitive(true);
+        {
+          add_document_menuitem->set_sensitive(true);
+          remove_gallery_menuitem->set_sensitive(true);
+          gallery_properties_menuitem->set_sensitive(true);
+        }
       else
-        gallery_properties_menuitem->set_sensitive(false);
+        {
+          gallery_properties_menuitem->set_sensitive(false);
+          add_document_menuitem->set_sensitive(false);
+          remove_gallery_menuitem->set_sensitive(false);
+        }
+      add_gallery_menuitem->set_sensitive(true);
+      Document *document = get_selected_document();
+      if (item_tagging_widget && show_tags_menuitem->get_active() == true)
+        {
+          if (gallery || document)
+            item_tagging_widget->get_widget()->set_sensitive(true);
+          else
+            item_tagging_widget->get_widget()->set_sensitive(false);
+        }
     }
   else
     {
@@ -806,12 +875,26 @@ void TapiocaWindow::update_menuitems()
       gallery_properties_menuitem->set_sensitive(false);
       export_menuitem->set_sensitive(false);
       contents->set_sensitive(false);
+      add_document_menuitem->set_sensitive(false);
+      remove_document_menuitem->set_sensitive(false);
+      add_gallery_menuitem->set_sensitive(false);
+      remove_gallery_menuitem->set_sensitive(false);
+      save_toolbutton->set_sensitive(false);
+      publish_toolbutton->set_sensitive(false);
     }
 
   if (get_selected_document() != NULL)
-    image_properties_menuitem->set_sensitive(true);
+    {
+      image_properties_menuitem->set_sensitive(true);
+      remove_document_menuitem->set_sensitive(true);
+      view_document_menuitem->set_sensitive(true);
+    }
   else
-    image_properties_menuitem->set_sensitive(false);
+    {
+      image_properties_menuitem->set_sensitive(false);
+      remove_document_menuitem->set_sensitive(false);
+      view_document_menuitem->set_sensitive(false);
+    }
   update_window_title();
 }
 
@@ -824,11 +907,27 @@ void TapiocaWindow::on_gallery_selected()
     gallery_label->set_text(gallery->get_title());
   if (gallery && gallery->size() > 0)
     iconview->scroll_to_path(Gtk::TreeModel::Path("0"), false, 0.0, 0.0);
+  if (show_tags_menuitem->get_active() == true && item_tagging_widget &&
+      session)
+    {
+      Document *document = get_selected_document();
+      if (!document)
+        item_tagging_widget->fill_tags(gallery->get_tags(), 
+                                       session->get_tags());
+      else
+        item_tagging_widget->fill_tags(document->get_tags(), 
+                                       session->get_tags());
+    }
   update_menuitems();
 }
 
 void TapiocaWindow::on_document_selected()
 {
+  Document *document = get_selected_document();
+  if (show_tags_menuitem->get_active() == true && document && 
+      item_tagging_widget && session)
+    item_tagging_widget->fill_tags(document->get_tags(), 
+                                   session->get_tags());
   update_menuitems();
 }
 
@@ -892,8 +991,18 @@ void TapiocaWindow::add_document(Tapioca::Document *document)
       
 void TapiocaWindow::on_treeview_clicked(GdkEventButton* event)
 {
-  if (event->type == GDK_BUTTON_PRESS &&event->button == 3)
+  if (event->type == GDK_BUTTON_PRESS && event->button == 3)
     {
+      Gtk::Menu::MenuList& menulist = treeview_context_menu.items();
+      Gallery *gallery = get_selected_gallery();
+      //Properties, and Remove theorhetically can be turned off.
+      //but right clicking the treeview causes a gallery to become selected.
+      //...so they can't actually be turned off.
+      Gtk::Menu::MenuList::iterator i = menulist.begin();
+      (*i).set_sensitive(gallery != NULL);
+      i++;
+      i++;
+      (*i).set_sensitive(gallery != NULL);
       treeview_context_menu.popup(event->button, event->time);
     }
 }
@@ -907,11 +1016,16 @@ void TapiocaWindow::on_iconview_clicked(GdkEventButton* event)
       if (path.empty() == false)
         iconview->select_path(path);
       Document *document = get_selected_document();
-      //View and Properties can be turned off.
+      Gallery *gallery = get_selected_gallery();
+      //View, Properties, Add and Remove can be turned off.
       Gtk::Menu::MenuList::iterator i = menulist.begin();
       (*i).set_sensitive(document != NULL);
       i++;
       (*i).set_sensitive(document != NULL);
+      i++;
+      (*i).set_sensitive(gallery != NULL);
+      i++;
+      (*i).set_sensitive(document != NULL);
       iconview_context_menu.popup(event->button, event->time);
     }
 }
@@ -1137,35 +1251,38 @@ void TapiocaWindow::update_window_title()
   set_title(title);
 }
 
+void TapiocaWindow::view_document(Document *document)
+{
+  if (document->is_local() == true)
+    view_downloaded_document(document, document->get_filename());
+  else
+    {
+      //short circuit: did we already download it?
+      if (document->get_original_filename() != "")
+        {
+          std::string file = temporary_directory + "/" + document->get_export_filename();
+          if (Document::exists(file) == true)
+            {
+              view_downloaded_document(document, file);
+              return;
+            }
+        }
+
+      bool have_credentials = true;
+      if (session->get_profile()->get_username() == "" ||
+          session->get_profile()->get_password() == "")
+        have_credentials = false;
+      on_download_document_attempted
+        (have_credentials, document, 
+         sigc::mem_fun(this, &TapiocaWindow::view_downloaded_document));
+    }
+}
+
 void TapiocaWindow::on_view_document_clicked()
 {
   Document *document = get_selected_document();
   if (document)
-    {
-      if (document->is_local() == true)
-        view_downloaded_document(document, document->get_filename());
-      else
-        {
-          //short circuit: did we already download it?
-          if (document->get_original_filename() != "")
-            {
-              std::string file = temporary_directory + "/" + document->get_export_filename();
-              if (Document::exists(file) == true)
-                {
-                  view_downloaded_document(document, file);
-                  return;
-                }
-            }
-            
-          bool have_credentials = true;
-          if (session->get_profile()->get_username() == "" ||
-              session->get_profile()->get_password() == "")
-            have_credentials = false;
-          on_download_document_attempted
-            (have_credentials, document, 
-             sigc::mem_fun(this, &TapiocaWindow::view_downloaded_document));
-        }
-    }
+    view_document(document);
 }
 
 void TapiocaWindow::on_document_downloaded_progress(Document *document, ProgressDialog &d)
@@ -1426,7 +1543,7 @@ void TapiocaWindow::on_pulled_documents_for_gallery(Gallery *gallery, std::list<
   s->pull_documents_for_gallery(g,sigc::bind<std::list<Gallery*>, Session *, ProgressDialog &>(sigc::mem_fun(*this, &TapiocaWindow::on_pulled_documents_for_gallery), galleries, s, d), thumbnail_urls);
 }
 
-void TapiocaWindow::add_published_galleries(sigc::slot<void> slot)
+void TapiocaWindow::add_published_galleries(sigc::slot<void> slot, ProgressDialog &d)
 {
   if (publish_changes.added_galleries.size() == 0)
     {
@@ -1440,10 +1557,10 @@ void TapiocaWindow::add_published_galleries(sigc::slot<void> slot)
       (gallery, sigc::bind(sigc::mem_fun
                            (*this, 
                             &TapiocaWindow::on_published_gallery_added), 
-                           slot));
+                           slot, d));
 }
 
-void TapiocaWindow::remove_published_galleries(sigc::slot<void> slot)
+void TapiocaWindow::remove_published_galleries(sigc::slot<void> slot, ProgressDialog &d)
 {
   if (publish_changes.removed_documents.size() == 0)
     {
@@ -1458,26 +1575,30 @@ void TapiocaWindow::remove_published_galleries(sigc::slot<void> slot)
       (gallery, sigc::bind(sigc::mem_fun
                            (*this, 
                             &TapiocaWindow::on_published_gallery_removed), 
-                           slot));
+                           slot, d));
 }
 
-void TapiocaWindow::publish(SessionChanges &changes)
+void TapiocaWindow::publish(SessionChanges &changes, ProgressDialog &d)
 {
   publish_changes = changes;
   //kick it off
-  add_published_galleries(sigc::mem_fun(*this, &TapiocaWindow::on_published_galleries_added));
+  //we chain the phases together.
+  //first we add galleries
+  d.pulse();
+  add_published_galleries(sigc::bind<ProgressDialog&>(sigc::mem_fun(*this, &TapiocaWindow::on_published_galleries_added), d), d);
 
 }
 
-void TapiocaWindow::on_published_galleries_added()
+void TapiocaWindow::on_published_galleries_added(ProgressDialog &d)
 {
   //now it's time to do the move documents.
+  d.pulse();
   publish_changes.find_moved_docs();
   move_published_documents
-    (sigc::mem_fun(*this, &TapiocaWindow::on_published_documents_moved));
+    (sigc::bind(sigc::mem_fun(*this, &TapiocaWindow::on_published_documents_moved), d), d);
 }
 
-void TapiocaWindow::move_published_documents(sigc::slot<void> slot)
+void TapiocaWindow::move_published_documents(sigc::slot<void> slot, ProgressDialog &d)
 {
   if (publish_changes.moved_documents.size() == 0)
     {
@@ -1493,14 +1614,17 @@ void TapiocaWindow::move_published_documents(sigc::slot<void> slot)
       session->move_document
         (src, dest, sigc::bind
          (sigc::mem_fun(this, &TapiocaWindow::on_published_document_moved), 
-          slot));
+          slot, d));
     }
 }
 
-void TapiocaWindow::on_published_document_moved(Document *src, Document *dest, sigc::slot<void> slot)
+void TapiocaWindow::on_published_document_moved(Document *src, Document *dest, sigc::slot<void> slot, ProgressDialog &d)
 {
+  d.pulse();
   publish_changes.moved_documents.erase
     (publish_changes.moved_documents.begin());
+  d.set_text(String::ucompose("%1 documents left to move", 
+                              publish_changes.moved_documents.size()));
 
   //Q. what do we do when the doc was moved?
   //A. we get rid of the source document, and update the new document's id.
@@ -1510,28 +1634,31 @@ void TapiocaWindow::on_published_document_moved(Document *src, Document *dest, s
     gallery->remove_document(src);
   delete src;
   if (publish_changes.moved_documents.size() > 0)
-    move_published_documents(slot);
+    move_published_documents(slot, d);
   else
     slot();
 }
 
-void TapiocaWindow::on_published_documents_moved()
+void TapiocaWindow::on_published_documents_moved(ProgressDialog &d)
 {
-  remove_published_galleries(sigc::mem_fun(*this, &TapiocaWindow::on_published_galleries_removed));
+  d.pulse();
+  remove_published_galleries(sigc::bind(sigc::mem_fun(*this, &TapiocaWindow::on_published_galleries_removed), d), d);
   //now we remove galleries.
 }
 
-void TapiocaWindow::on_published_galleries_removed()
+void TapiocaWindow::on_published_galleries_removed(ProgressDialog &d)
 {
-  add_published_documents(sigc::mem_fun(*this, &TapiocaWindow::on_published_documents_added));
+  d.pulse();
+  add_published_documents(sigc::bind(sigc::mem_fun(*this, &TapiocaWindow::on_published_documents_added), d), d);
 }
 
-void TapiocaWindow::on_published_documents_added()
+void TapiocaWindow::on_published_documents_added(ProgressDialog &d)
 {
-  remove_published_documents(sigc::mem_fun(*this, &TapiocaWindow::on_published_documents_removed));
+  d.pulse();
+  remove_published_documents(sigc::bind(sigc::mem_fun(*this, &TapiocaWindow::on_published_documents_removed), d), d);
 }
 
-void TapiocaWindow::add_published_documents(sigc::slot<void> slot)
+void TapiocaWindow::add_published_documents(sigc::slot<void> slot, ProgressDialog &d)
 {
   if (publish_changes.added_documents.size() == 0)
     {
@@ -1545,10 +1672,10 @@ void TapiocaWindow::add_published_documents(sigc::slot<void> slot)
       (document, sigc::bind(sigc::mem_fun
                            (*this, 
                             &TapiocaWindow::on_published_document_added), 
-                           slot));
+                           slot, d));
 }
 
-void TapiocaWindow::remove_published_documents(sigc::slot<void> slot)
+void TapiocaWindow::remove_published_documents(sigc::slot<void> slot, ProgressDialog &d)
 {
   if (publish_changes.removed_documents.size() == 0)
     {
@@ -1562,40 +1689,47 @@ void TapiocaWindow::remove_published_documents(sigc::slot<void> slot)
       (document, sigc::bind(sigc::mem_fun
                            (*this, 
                             &TapiocaWindow::on_published_document_removed), 
-                           slot));
+                           slot, d));
 }
 
-void TapiocaWindow::on_published_document_removed(Document *document, sigc::slot<void> slot)
+void TapiocaWindow::on_published_document_removed(Document *document, sigc::slot<void> slot, ProgressDialog &d)
 {
+  d.pulse();
   publish_changes.removed_documents.remove
     (publish_changes.removed_documents.front());
+  d.set_text(String::ucompose("%1 documents left to remove", 
+                              publish_changes.removed_documents.size()));
   Gallery *gallery = 
     session->get_profile()->find_by_id(document->get_gallery_id());
   gallery->remove_document(document);
   delete document;
   if (publish_changes.removed_documents.size() > 0)
-    remove_published_documents(slot);
+    remove_published_documents(slot, d);
   else
     slot();
 }
 
-void TapiocaWindow::on_published_document_added(Document *document, sigc::slot<void> slot)
+void TapiocaWindow::on_published_document_added(Document *document, sigc::slot<void> slot, ProgressDialog &d)
 {
+  d.pulse();
   publish_changes.added_documents.remove
     (publish_changes.added_documents.front());
+  d.set_text(String::ucompose("%1 documents left to add", 
+                              publish_changes.added_documents.size()));
   if (publish_changes.added_documents.size() > 0)
-    add_published_documents(slot);
+    add_published_documents(slot, d);
   else
     slot();
 }
 
-void TapiocaWindow::on_published_documents_removed()
+void TapiocaWindow::on_published_documents_removed(ProgressDialog &d)
 {
+  d.pulse();
   //now time for modified galleries, and modified documents
-  modify_published_galleries(sigc::mem_fun(*this, &TapiocaWindow::on_published_galleries_modified));
+  modify_published_galleries(sigc::bind(sigc::mem_fun(*this, &TapiocaWindow::on_published_galleries_modified), d), d);
 }
 
-void TapiocaWindow::modify_published_galleries(sigc::slot<void> slot)
+void TapiocaWindow::modify_published_galleries(sigc::slot<void> slot, ProgressDialog &d)
 {
   if (publish_changes.modified_galleries.size() == 0)
     {
@@ -1609,25 +1743,29 @@ void TapiocaWindow::modify_published_galleries(sigc::slot<void> slot)
       (gallery, sigc::bind(sigc::mem_fun
                            (*this, 
                             &TapiocaWindow::on_published_gallery_modified), 
-                           slot));
+                           slot, d));
 }
 
-void TapiocaWindow::on_published_gallery_modified(Gallery *gallery, sigc::slot<void> slot)
+void TapiocaWindow::on_published_gallery_modified(Gallery *gallery, sigc::slot<void> slot, ProgressDialog &d)
 {
+  d.pulse();
   publish_changes.modified_galleries.remove
     (publish_changes.modified_galleries.front());
+  d.set_text(String::ucompose("%1 galleries left to modify", 
+                              publish_changes.modified_galleries.size()));
   if (publish_changes.modified_galleries.size() > 0)
-    modify_published_galleries(slot);
+    modify_published_galleries(slot, d);
   else
     slot();
 }
 
-void TapiocaWindow::on_published_galleries_modified()
+void TapiocaWindow::on_published_galleries_modified(ProgressDialog &d)
 {
-  modify_published_documents(sigc::mem_fun(*this, &TapiocaWindow::on_published_documents_modified));
+  d.pulse();
+  modify_published_documents(sigc::bind(sigc::mem_fun(*this, &TapiocaWindow::on_published_documents_modified), d), d);
 }
 
-void TapiocaWindow::modify_published_documents(sigc::slot<void> slot)
+void TapiocaWindow::modify_published_documents(sigc::slot<void> slot, ProgressDialog &d)
 {
   if (publish_changes.modified_docs.size() == 0)
     {
@@ -1641,19 +1779,123 @@ void TapiocaWindow::modify_published_documents(sigc::slot<void> slot)
       (document, sigc::bind(sigc::mem_fun
                            (*this, 
                             &TapiocaWindow::on_published_document_modified), 
-                           slot));
+                           slot, d));
 }
 
-void TapiocaWindow::on_published_document_modified(Document *document, sigc::slot<void> slot)
+void TapiocaWindow::on_published_document_modified(Document *document, sigc::slot<void> slot, ProgressDialog &d)
 {
+  d.pulse();
   publish_changes.modified_docs.remove
     (publish_changes.modified_docs.front());
+  d.set_text(String::ucompose("%1 documents left to modify", 
+                              publish_changes.modified_docs.size()));
   if (publish_changes.modified_docs.size() > 0)
-    modify_published_documents(slot);
+    modify_published_documents(slot, d);
   else
     slot();
 }
 
-void TapiocaWindow::on_published_documents_modified()
+void TapiocaWindow::on_published_documents_modified(ProgressDialog &d)
 {
+  //and we're done with the giant chain of publishing phases.
+  d.hide();
+}
+
+void TapiocaWindow::on_fullscreen_toggled()
+{
+  if (fullscreen_menuitem->get_active() == true)
+    get_window()->maximize();
+  else
+    get_window()->unmaximize();
+}
+
+void TapiocaWindow::on_show_tags_toggled()
+{
+  if (show_tags_menuitem->get_active() == true)
+    show_tagging_widget();
+  else
+    unshow_tagging_widget();
+}
+
+void TapiocaWindow::show_tagging_widget()
+{
+  if (item_tagging_widget != NULL)
+    delete item_tagging_widget;
+  item_tagging_widget = new ItemTaggingWidget();
+  item_tagging_widget->tags_changed.connect
+    (sigc::mem_fun(*this, &TapiocaWindow::on_tags_changed));
+  Gtk::VBox *item_tagging_widget_container = new Gtk::VBox();
+  contents->pack_start(*item_tagging_widget_container, true, true, 0);
+  Gtk::Label *label = new Gtk::Label();
+  label->set_markup("<b>Tags</b>");
+  Gtk::Button *button = new Gtk::Button();
+  Gtk::Image *image = new Gtk::Image(Gtk::StockID(Gtk::Stock::CLOSE), 
+                                     Gtk::ICON_SIZE_MENU);
+  button->add(*image);
+  button->signal_clicked().connect
+    (sigc::bind(sigc::mem_fun(show_tags_menuitem, &Gtk::CheckMenuItem::set_active), false));
+
+  Gtk::HBox *hbox = new Gtk::HBox();
+  Gtk::VSeparator *sep = new Gtk::VSeparator();
+  hbox->pack_start(*sep, false, false, 0);
+  hbox->pack_start(*label, true, false, 0);
+  hbox->pack_start(*button, false, false, 0);
+  item_tagging_widget_container->pack_start(*hbox, false, false, 0);
+  Gtk::HSeparator *hsep = new Gtk::HSeparator();
+  item_tagging_widget_container->pack_start(*hsep, false, false, 0);
+  item_tagging_widget->get_widget()->reparent(*item_tagging_widget_container);
+  image->show_all();
+  if (session)
+    {
+      std::list<std::string> tags = session->get_tags();
+      Document *document = get_selected_document();
+      if (!document)
+        {
+          Gallery *gallery = get_selected_gallery();
+          if (gallery)
+            item_tagging_widget->fill_tags(gallery->get_tags(), tags);
+        }
+      else
+        item_tagging_widget->fill_tags(document->get_tags(), tags);
+    }
+  contents->show_all();
+}
+
+void TapiocaWindow::unshow_tagging_widget()
+{
+  std::list<Gtk::Widget*> widgets = contents->get_children();
+  contents->remove(*widgets.back());
+  contents->show_all();
+}
+
+void TapiocaWindow::on_tags_changed(std::list<std::string> tags)
+{
+  Document *document = get_selected_document();
+  Gallery *gallery = get_selected_gallery();
+  if (document)
+    document->set_tags(tags);
+  else
+    {
+      if (gallery)
+        gallery->set_tags(tags);
+    }
+}
+
+void TapiocaWindow::on_gallery_tags_changed(std::list<std::string> tags, Tapioca::Gallery *gallery)
+{
+  if (gallery->get_id() == get_selected_gallery()->get_id() && 
+      get_selected_document() == NULL)
+    {
+      on_tags_changed(tags);
+      item_tagging_widget->fill_tags(tags, session->get_tags());
+    }
+}
+
+void TapiocaWindow::on_document_tags_changed(std::list<std::string> tags, Tapioca::Document* document)
+{
+  if (document->get_id() == get_selected_document()->get_id())
+    {
+      on_tags_changed(tags);
+      item_tagging_widget->fill_tags(tags, session->get_tags());
+    }
 }
